# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import testtools

from kmip.core import attributes
from kmip.core import enums
from kmip.core import misc
from kmip.core import objects
from kmip.core import secrets
from kmip.core import utils

from kmip.core.messages import payloads


class TestGetRequestPayload(testtools.TestCase):
    """
    Test suite for the Get request payload.
    """

    def setUp(self):
        super(TestGetRequestPayload, self).setUp()

        # Encoding obtained in part from the KMIP 1.1 testing document,
        # Sections 3.1.3 and 14.1. The rest of the encoding was built by
        # hand.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Unique Identifier - 49a1ca88-6bea-4fb2-b450-7e58802c3038
        #     Key Format Type - Raw
        #     Key Compression Type - EC Public Key Type Uncompressed
        #     Key Wrapping Specification
        #         Key Wrapping Method - Encrypt
        #         Encryption Key Information
        #             Unique Identifier - 100182d5-72b8-47aa-8383-4d97d512e98a
        #             Cryptographic Parameters
        #                 Block Cipher Mode - NIST Key Wrap
        #         Encoding Option - No Encoding

        self.full_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\xC8'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x34\x39\x61\x31\x63\x61\x38\x38\x2D\x36\x62\x65\x61\x2D\x34\x66'
            b'\x62\x32\x2D\x62\x34\x35\x30\x2D\x37\x65\x35\x38\x38\x30\x32\x63'
            b'\x33\x30\x33\x38\x00\x00\x00\x00'
            b'\x42\x00\x42\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x41\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x47\x01\x00\x00\x00\x70'
            b'\x42\x00\x9E\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x36\x01\x00\x00\x00\x48'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x31\x30\x30\x31\x38\x32\x64\x35\x2D\x37\x32\x62\x38\x2D\x34\x37'
            b'\x61\x61\x2D\x38\x33\x38\x33\x2D\x34\x64\x39\x37\x64\x35\x31\x32'
            b'\x65\x39\x38\x61\x00\x00\x00\x00'
            b'\x42\x00\x2B\x01\x00\x00\x00\x10'
            b'\x42\x00\x11\x05\x00\x00\x00\x04\x00\x00\x00\x0D\x00\x00\x00\x00'
            b'\x42\x00\xA3\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 3.1.3.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Unique Identifier - 49a1ca88-6bea-4fb2-b450-7e58802c3038

        self.partial_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x30'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x34\x39\x61\x31\x63\x61\x38\x38\x2D\x36\x62\x65\x61\x2D\x34\x66'
            b'\x62\x32\x2D\x62\x34\x35\x30\x2D\x37\x65\x35\x38\x38\x30\x32\x63'
            b'\x33\x30\x33\x38\x00\x00\x00\x00'
        )

        self.empty_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x00'
        )

    def tearDown(self):
        super(TestGetRequestPayload, self).tearDown()

    def test_init(self):
        """
        Test that a Get request payload can be constructed with no arguments.
        """
        payload = payloads.GetRequestPayload()

        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

    def test_init_with_args(self):
        """
        Test that a Get request payload can be constructed with valid values.
        """
        payload = payloads.GetRequestPayload(
            unique_identifier='00000000-2222-4444-6666-888888888888',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT
            )
        )

        self.assertEqual(
            '00000000-2222-4444-6666-888888888888',
            payload.unique_identifier
        )
        self.assertEqual(enums.KeyFormatType.RAW, payload.key_format_type)
        self.assertEqual(
            enums.KeyCompressionType.EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            payload.key_compression_type
        )
        self.assertIsInstance(
            payload.key_wrapping_specification,
            objects.KeyWrappingSpecification
        )
        self.assertEqual(
            enums.WrappingMethod.ENCRYPT,
            payload.key_wrapping_specification.wrapping_method
        )

    def test_invalid_unique_identifier(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the unique identifier of a Get request payload.
        """
        kwargs = {'unique_identifier': 0}
        self.assertRaisesRegex(
            TypeError,
            "Unique identifier must be a string.",
            payloads.GetRequestPayload,
            **kwargs
        )

        args = (payloads.GetRequestPayload(), 'unique_identifier', 0)
        self.assertRaisesRegex(
            TypeError,
            "Unique identifier must be a string.",
            setattr,
            *args
        )

    def test_invalid_key_format_type(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the key format type of a Get request payload.
        """
        kwargs = {'key_format_type': 'invalid'}
        self.assertRaisesRegex(
            TypeError,
            "Key format type must be a KeyFormatType enumeration.",
            payloads.GetRequestPayload,
            **kwargs
        )

        args = (payloads.GetRequestPayload(), 'key_format_type', 'invalid')
        self.assertRaisesRegex(
            TypeError,
            "Key format type must be a KeyFormatType enumeration.",
            setattr,
            *args
        )

    def test_invalid_key_compression_type(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the key compression type of a Get request payload.
        """
        kwargs = {'key_compression_type': 'invalid'}
        self.assertRaisesRegex(
            TypeError,
            "Key compression type must be a KeyCompressionType enumeration.",
            payloads.GetRequestPayload,
            **kwargs
        )

        args = (
            payloads.GetRequestPayload(),
            'key_compression_type',
            'invalid'
        )
        self.assertRaisesRegex(
            TypeError,
            "Key compression type must be a KeyCompressionType enumeration.",
            setattr,
            *args
        )

    def test_invalid_key_wrapping_specification(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the key wrapping specification of a Get request payload.
        """
        kwargs = {'key_wrapping_specification': 'invalid'}
        self.assertRaisesRegex(
            TypeError,
            "Key wrapping specification must be a KeyWrappingSpecification "
            "struct.",
            payloads.GetRequestPayload,
            **kwargs
        )

        args = (
            payloads.GetRequestPayload(),
            'key_wrapping_specification',
            'invalid'
        )
        self.assertRaisesRegex(
            TypeError,
            "Key wrapping specification must be a KeyWrappingSpecification "
            "struct.",
            setattr,
            *args
        )

    def test_read(self):
        """
        Test that a GetRequestPayload struct can be read from a data stream.
        """
        payload = payloads.GetRequestPayload()

        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

        payload.read(self.full_encoding)

        self.assertEqual(
            '49a1ca88-6bea-4fb2-b450-7e58802c3038',
            payload.unique_identifier
        )
        self.assertEqual(enums.KeyFormatType.RAW, payload.key_format_type)
        self.assertEqual(
            enums.KeyCompressionType.EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            payload.key_compression_type
        )
        self.assertIsInstance(
            payload.key_wrapping_specification,
            objects.KeyWrappingSpecification
        )
        k = payload.key_wrapping_specification
        self.assertEqual(
            enums.WrappingMethod.ENCRYPT,
            k.wrapping_method
        )
        self.assertIsInstance(
            k.encryption_key_information,
            objects.EncryptionKeyInformation
        )
        e = k.encryption_key_information
        self.assertEqual(
            '100182d5-72b8-47aa-8383-4d97d512e98a',
            e.unique_identifier
        )
        self.assertIsInstance(
            e.cryptographic_parameters,
            attributes.CryptographicParameters
        )
        self.assertEqual(
            enums.BlockCipherMode.NIST_KEY_WRAP,
            e.cryptographic_parameters.block_cipher_mode
        )
        self.assertEqual(
            k.encoding_option,
            enums.EncodingOption.NO_ENCODING
        )

    def test_read_partial(self):
        """
        Test that a GetRequestPayload struct can be read from a partial data
        stream.
        """
        payload = payloads.GetRequestPayload()

        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

        payload.read(self.partial_encoding)

        self.assertEqual(
            '49a1ca88-6bea-4fb2-b450-7e58802c3038',
            payload.unique_identifier
        )
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

    def test_read_empty(self):
        """
        Test that a GetRequestPayload struct can be read from an empty data
        stream.
        """
        payload = payloads.GetRequestPayload()

        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

        payload.read(self.empty_encoding)

        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.key_format_type)
        self.assertEqual(None, payload.key_compression_type)
        self.assertEqual(None, payload.key_wrapping_specification)

    def test_write(self):
        """
        Test that a GetRequestPayload struct can be written to a data stream.
        """
        payload = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )
        stream = utils.BytearrayStream()

        payload.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_partial(self):
        """
        Test that a partially defined GetRequestPayload struct can be written
        to a data stream.
        """
        payload = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        stream = utils.BytearrayStream()

        payload.write(stream)

        self.assertEqual(len(self.partial_encoding), len(stream))
        self.assertEqual(str(self.partial_encoding), str(stream))

    def test_write_empty(self):
        """
        Test that an empty GetRequestPayload struct can be written to a data
        stream.
        """
        payload = payloads.GetRequestPayload()
        stream = utils.BytearrayStream()

        payload.write(stream)

        self.assertEqual(len(self.empty_encoding), len(stream))
        self.assertEqual(str(self.empty_encoding), str(stream))

    def test_equal_on_equal(self):
        """
        Test that the equality operator returns True when comparing two
        GetRequestPayload structs with the same data.
        """
        a = payloads.GetRequestPayload()
        b = payloads.GetRequestPayload()

        self.assertTrue(a == b)
        self.assertTrue(b == a)

        a = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )
        b = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_not_equal_unique_identifier(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different unique identifiers.
        """
        a = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        b = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c303f'
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_key_format_type(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different key format types.
        """
        a = payloads.GetRequestPayload(
            key_format_type=enums.KeyFormatType.RAW
        )
        b = payloads.GetRequestPayload(
            key_format_type=enums.KeyFormatType.OPAQUE
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_key_compression_type(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different key compression types.
        """
        a = payloads.GetRequestPayload(
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED
        )
        b = payloads.GetRequestPayload(
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_X9_62_HYBRID
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_key_wrapping_specification(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different key wrapping specifications.
        """
        a = payloads.GetRequestPayload(
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT_THEN_MAC_SIGN,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-ffff-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )
        b = payloads.GetRequestPayload(
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_type_mismatch(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different types.
        """
        a = payloads.GetRequestPayload()
        b = 'invalid'

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_not_equal_on_equal(self):
        """
        Test that the inequality operator returns False when comparing two
        GetRequestPayload structs with the same data.
        """
        a = payloads.GetRequestPayload()
        b = payloads.GetRequestPayload()

        self.assertFalse(a != b)
        self.assertFalse(b != a)

        a = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )
        b = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_not_equal_unique_identifier(self):
        """
        Test that the inequality operator returns True when comparing two
        GetRequestPayload structs with different unique identifiers.
        """
        a = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        b = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c303f'
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_key_format_type(self):
        """
        Test that the inequality operator returns True when comparing two
        GetRequestPayload structs with different key format types.
        """
        a = payloads.GetRequestPayload(
            key_format_type=enums.KeyFormatType.RAW
        )
        b = payloads.GetRequestPayload(
            key_format_type=enums.KeyFormatType.OPAQUE
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_key_compression_type(self):
        """
        Test that the equality operator returns False when comparing two
        GetRequestPayload structs with different key compression types.
        """
        a = payloads.GetRequestPayload(
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED
        )
        b = payloads.GetRequestPayload(
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_X9_62_HYBRID
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_key_wrapping_specification(self):
        """
        Test that the inequality operator returns True when comparing two
        GetRequestPayload structs with different key wrapping specifications.
        """
        a = payloads.GetRequestPayload(
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT_THEN_MAC_SIGN,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-ffff-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )
        b = payloads.GetRequestPayload(
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_type_mismatch(self):
        """
        Test that the inequality operator returns True when comparing two
        GetRequestPayload structs with different types.
        """
        a = payloads.GetRequestPayload()
        b = 'invalid'

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_repr(self):
        """
        Test that repr can be applied to a GetRequestPayload struct.
        """
        payload = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        expected = (
            "GetRequestPayload("
            "unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038', "
            "key_format_type=KeyFormatType.RAW, "
            "key_compression_type="
            "KeyCompressionType.EC_PUBLIC_KEY_TYPE_UNCOMPRESSED, "
            "key_wrapping_specification="
            "KeyWrappingSpecification("
            "wrapping_method=WrappingMethod.ENCRYPT, "
            "encryption_key_information=EncryptionKeyInformation("
            "unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a', "
            "cryptographic_parameters=CryptographicParameters("
            "block_cipher_mode=BlockCipherMode.NIST_KEY_WRAP, "
            "padding_method=None, "
            "hashing_algorithm=None, "
            "key_role_type=None, "
            "digital_signature_algorithm=None, "
            "cryptographic_algorithm=None, "
            "random_iv=None, "
            "iv_length=None, "
            "tag_length=None, "
            "fixed_field_length=None, "
            "invocation_field_length=None, "
            "counter_length=None, "
            "initial_counter_value=None)), "
            "mac_signature_key_information=None, "
            "attribute_names=None, "
            "encoding_option=EncodingOption.NO_ENCODING))"
        )
        observed = repr(payload)

        self.assertEqual(expected, observed)

    def test_str(self):
        """
        Test that str can be applied to a GetRequestPayload struct.
        """
        payload = payloads.GetRequestPayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            key_format_type=enums.KeyFormatType.RAW,
            key_compression_type=enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            key_wrapping_specification=objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        )

        expected = str({
            'unique_identifier': '49a1ca88-6bea-4fb2-b450-7e58802c3038',
            'key_format_type': enums.KeyFormatType.RAW,
            'key_compression_type': enums.KeyCompressionType.
            EC_PUBLIC_KEY_TYPE_UNCOMPRESSED,
            'key_wrapping_specification': objects.KeyWrappingSpecification(
                wrapping_method=enums.WrappingMethod.ENCRYPT,
                encryption_key_information=objects.EncryptionKeyInformation(
                    unique_identifier='100182d5-72b8-47aa-8383-4d97d512e98a',
                    cryptographic_parameters=attributes.
                    CryptographicParameters(
                        block_cipher_mode=enums.BlockCipherMode.NIST_KEY_WRAP
                    )
                ),
                encoding_option=enums.EncodingOption.NO_ENCODING
            )
        })
        observed = str(payload)

        self.assertEqual(expected, observed)


class TestGetResponsePayload(testtools.TestCase):
    """
    Test suite for the Get response payload.
    """

    def setUp(self):
        super(TestGetResponsePayload, self).setUp()

        # Encoding obtained from the KMIP 1.1 testing document, Section 3.1.3.
        #
        # This encoding matches the following set of values:
        # Response Payload
        #     Object Type - Symmetric Key
        #     Unique Identifier - 49a1ca88-6bea-4fb2-b450-7e58802c3038
        #     Symmetric Key
        #         Key Block
        #             Key Format Type - Raw
        #             Key Value
        #                 Key Material - 0x7367578051012A6D134A855E25C8CD5E4C
        #                                A131455729D3C8
        #             Cryptographic Algorithm - 3DES
        #             Cryptographic Length - 168

        self.full_encoding = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\xA8'
            b'\x42\x00\x57\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x34\x39\x61\x31\x63\x61\x38\x38\x2D\x36\x62\x65\x61\x2D\x34\x66'
            b'\x62\x32\x2D\x62\x34\x35\x30\x2D\x37\x65\x35\x38\x38\x30\x32\x63'
            b'\x33\x30\x33\x38\x00\x00\x00\x00'
            b'\x42\x00\x8F\x01\x00\x00\x00\x60'
            b'\x42\x00\x40\x01\x00\x00\x00\x58'
            b'\x42\x00\x42\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x45\x01\x00\x00\x00\x20'
            b'\x42\x00\x43\x08\x00\x00\x00\x18'
            b'\x73\x67\x57\x80\x51\x01\x2A\x6D\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
            b'\x42\x00\x28\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x2A\x02\x00\x00\x00\x04\x00\x00\x00\xA8\x00\x00\x00\x00'
        )

        self.partial_encoding_missing_object_type = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\xA0'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x34\x39\x61\x31\x63\x61\x38\x38\x2D\x36\x62\x65\x61\x2D\x34\x66'
            b'\x62\x32\x2D\x62\x34\x35\x30\x2D\x37\x65\x35\x38\x38\x30\x32\x63'
            b'\x33\x30\x33\x38\x00\x00\x00\x00'
            b'\x42\x00\x8F\x01\x00\x00\x00\x60'
            b'\x42\x00\x40\x01\x00\x00\x00\x58'
            b'\x42\x00\x42\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x45\x01\x00\x00\x00\x20'
            b'\x42\x00\x43\x08\x00\x00\x00\x18'
            b'\x73\x67\x57\x80\x51\x01\x2A\x6D\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
            b'\x42\x00\x28\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x2A\x02\x00\x00\x00\x04\x00\x00\x00\xA8\x00\x00\x00\x00'
        )
        self.partial_encoding_missing_unique_id = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x78'
            b'\x42\x00\x57\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x8F\x01\x00\x00\x00\x60'
            b'\x42\x00\x40\x01\x00\x00\x00\x58'
            b'\x42\x00\x42\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x45\x01\x00\x00\x00\x20'
            b'\x42\x00\x43\x08\x00\x00\x00\x18'
            b'\x73\x67\x57\x80\x51\x01\x2A\x6D\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
            b'\x42\x00\x28\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x2A\x02\x00\x00\x00\x04\x00\x00\x00\xA8\x00\x00\x00\x00'
        )
        self.partial_encoding_missing_secret = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x40'
            b'\x42\x00\x57\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x34\x39\x61\x31\x63\x61\x38\x38\x2D\x36\x62\x65\x61\x2D\x34\x66'
            b'\x62\x32\x2D\x62\x34\x35\x30\x2D\x37\x65\x35\x38\x38\x30\x32\x63'
            b'\x33\x30\x33\x38\x00\x00\x00\x00'
        )

    def tearDown(self):
        super(TestGetResponsePayload, self).tearDown()

    def test_init(self):
        """
        Test that a GetRequestPayload struct can be constructed with no
        arguments.
        """
        payload = payloads.GetResponsePayload()

        self.assertEqual(None, payload.object_type)
        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.secret)

    def test_init_with_args(self):
        """
        Test that a GetRequestPayload struct can be constructed with valid
        values.
        """
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='11111111-3333-5555-7777-999999999999',
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )

        self.assertEqual(
            enums.ObjectType.SYMMETRIC_KEY,
            payload.object_type
        )
        self.assertEqual(
            '11111111-3333-5555-7777-999999999999',
            payload.unique_identifier
        )
        self.assertIsInstance(payload.secret, secrets.SymmetricKey)
        self.assertIsInstance(payload.secret.key_block, objects.KeyBlock)
        self.assertIsInstance(
            payload.secret.key_block.key_format_type,
            misc.KeyFormatType
        )
        self.assertEqual(
            enums.KeyFormatType.RAW,
            payload.secret.key_block.key_format_type.value
        )
        self.assertIsInstance(
            payload.secret.key_block.key_value,
            objects.KeyValue
        )
        self.assertIsInstance(
            payload.secret.key_block.key_value.key_material,
            objects.KeyMaterial
        )
        self.assertEqual(
            (
                b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
            ),
            payload.secret.key_block.key_value.key_material.value
        )
        self.assertIsInstance(
            payload.secret.key_block.cryptographic_algorithm,
            attributes.CryptographicAlgorithm
        )
        self.assertEqual(
            enums.CryptographicAlgorithm.TRIPLE_DES,
            payload.secret.key_block.cryptographic_algorithm.value
        )
        self.assertIsInstance(
            payload.secret.key_block.cryptographic_length,
            attributes.CryptographicLength
        )
        self.assertEqual(
            168,
            payload.secret.key_block.cryptographic_length.value
        )

    def test_invalid_object_type(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the object type of a GetResponsePayload struct.
        """
        kwargs = {'object_type': 'invalid'}
        self.assertRaisesRegex(
            TypeError,
            "Object type must be an ObjectType enumeration.",
            payloads.GetResponsePayload,
            **kwargs
        )

        args = (payloads.GetResponsePayload(), 'object_type', 'invalid')
        self.assertRaisesRegex(
            TypeError,
            "Object type must be an ObjectType enumeration.",
            setattr,
            *args
        )

    def test_invalid_unique_identifier(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the unique identifier of a GetResponsePayload struct.
        """
        kwargs = {'unique_identifier': 0}
        self.assertRaisesRegex(
            TypeError,
            "Unique identifier must be a string.",
            payloads.GetResponsePayload,
            **kwargs
        )

        args = (payloads.GetResponsePayload(), 'unique_identifier', 0)
        self.assertRaisesRegex(
            TypeError,
            "Unique identifier must be a string.",
            setattr,
            *args
        )

    def test_invalid_secret(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the secret of a GetResponsePayload struct.
        """
        kwargs = {'secret': 0}
        self.assertRaisesRegex(
            TypeError,
            "Secret must be one of the following structs: Certificate, "
            "OpaqueObject, PrivateKey, PublicKey, SecretData, SplitKey, "
            "SymmetricKey, Template",
            payloads.GetResponsePayload,
            **kwargs
        )

        args = (payloads.GetResponsePayload(), 'secret', 0)
        self.assertRaisesRegex(
            TypeError,
            "Secret must be one of the following structs: Certificate, "
            "OpaqueObject, PrivateKey, PublicKey, SecretData, SplitKey, "
            "SymmetricKey, Template",
            setattr,
            *args
        )

    def test_read(self):
        """
        Test that a GetResponsePayload struct can be read from a data stream.
        """
        payload = payloads.GetResponsePayload()

        self.assertEqual(None, payload.object_type)
        self.assertEqual(None, payload.unique_identifier)
        self.assertEqual(None, payload.secret)

        payload.read(self.full_encoding)

        self.assertEqual(enums.ObjectType.SYMMETRIC_KEY, payload.object_type)
        self.assertEqual(
            '49a1ca88-6bea-4fb2-b450-7e58802c3038',
            payload.unique_identifier
        )
        self.assertIsInstance(payload.secret, secrets.SymmetricKey)
        self.assertIsInstance(payload.secret.key_block, objects.KeyBlock)
        self.assertIsInstance(
            payload.secret.key_block.key_format_type,
            misc.KeyFormatType
        )
        self.assertEqual(
            enums.KeyFormatType.RAW,
            payload.secret.key_block.key_format_type.value
        )
        self.assertIsInstance(
            payload.secret.key_block.key_value,
            objects.KeyValue
        )
        self.assertIsInstance(
            payload.secret.key_block.key_value.key_material,
            objects.KeyMaterial
        )
        self.assertEqual(
            (
                b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
            ),
            payload.secret.key_block.key_value.key_material.value
        )
        self.assertIsInstance(
            payload.secret.key_block.cryptographic_algorithm,
            attributes.CryptographicAlgorithm
        )
        self.assertEqual(
            enums.CryptographicAlgorithm.TRIPLE_DES,
            payload.secret.key_block.cryptographic_algorithm.value
        )
        self.assertIsInstance(
            payload.secret.key_block.cryptographic_length,
            attributes.CryptographicLength
        )
        self.assertEqual(
            168,
            payload.secret.key_block.cryptographic_length.value
        )

    def test_read_missing_object_type(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when decoding the struct.
        """
        payload = payloads.GetResponsePayload()
        args = (self.partial_encoding_missing_object_type, )
        self.assertRaisesRegex(
            ValueError,
            "Parsed payload encoding is missing the object type field.",
            payload.read,
            *args
        )

    def test_read_missing_unique_identifier(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when decoding the struct.
        """
        payload = payloads.GetResponsePayload()
        args = (self.partial_encoding_missing_unique_id, )
        self.assertRaisesRegex(
            ValueError,
            "Parsed payload encoding is missing the unique identifier field.",
            payload.read,
            *args
        )

    def test_read_missing_secret(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when decoding the struct.
        """
        payload = payloads.GetResponsePayload()
        args = (self.partial_encoding_missing_secret, )
        self.assertRaisesRegex(
            ValueError,
            "Parsed payload encoding is missing the secret field.",
            payload.read,
            *args
        )

    def test_write(self):
        """
        Test that a GetResponsePayload struct can be written to a data stream.
        """
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )
        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_missing_object_type(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when encoding the struct.
        """
        payload = payloads.GetResponsePayload()
        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            ValueError,
            "Payload is missing the object type field.",
            payload.write,
            *args
        )

    def test_write_missing_unique_identifier(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when encoding the struct.
        """
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY
        )
        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            ValueError,
            "Payload is missing the unique identifier field.",
            payload.write,
            *args
        )

    def test_write_missing_secret(self):
        """
        Test that a ValueError gets raised when a required GetResponsePayload
        field is missing when encoding the struct.
        """
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            ValueError,
            "Payload is missing the secret field.",
            payload.write,
            *args
        )

    def test_equal_on_equal(self):
        """
        Test that the equality operator returns True when comparing two
        GetResponsePayload structs with the same data.
        """
        a = payloads.GetResponsePayload()
        b = payloads.GetResponsePayload()

        self.assertTrue(a == b)
        self.assertTrue(b == a)

        # TODO (peter-hamilton): Update this once equality is supported for
        # SymmetricKeys.
        secret = secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )

        a = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secret
        )
        b = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secret
        )

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_not_equal_object_type(self):
        """
        Test that the equality operator returns False when comparing two
        GetResponsePayload structs with different object type fields.
        """
        a = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY
        )
        b = payloads.GetResponsePayload(
            object_type=enums.ObjectType.OPAQUE_DATA
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_unique_identifier(self):
        """
        Test that the equality operator returns False when comparing two
        GetResponsePayload structs with different unique identifier fields.
        """
        a = payloads.GetResponsePayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        b = payloads.GetResponsePayload(
            unique_identifier='49a1ca88-6bea-4fb2-ffff-7e58802c3038'
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_secrets(self):
        """
        Test that the equality operator returns False when comparing two
        GetResponsePayload structs with different secret fields.
        """
        # TODO (peter-hamilton): Update this test case once SymmetricKeys
        # support proper field-based equality.
        a = payloads.GetResponsePayload(
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )
        b = payloads.GetResponsePayload(
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_type_mismatch(self):
        """
        Test that the equality operators returns False when comparing two
        GetResponsePayload structs with different types.
        """
        a = payloads.GetResponsePayload()
        b = 'invalid'

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_not_equal_on_equal(self):
        """
        Test that the inequality operator returns False when comparing two
        GetResponsePayload structs with the same data.
        """
        a = payloads.GetResponsePayload()
        b = payloads.GetResponsePayload()

        self.assertFalse(a != b)
        self.assertFalse(b != a)

        # TODO (peter-hamilton): Update this once equality is supported for
        # SymmetricKeys.
        secret = secrets.SymmetricKey(
            key_block=objects.KeyBlock(
                key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
                key_value=objects.KeyValue(
                    key_material=objects.KeyMaterial(
                        b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                        b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                        b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                    )
                ),
                cryptographic_algorithm=attributes.CryptographicAlgorithm(
                    enums.CryptographicAlgorithm.TRIPLE_DES
                ),
                cryptographic_length=attributes.CryptographicLength(168)
            )
        )

        a = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secret
        )
        b = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secret
        )

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_not_equal_object_type(self):
        """
        Test that the inequality operator returns True when comparing two
        GetResponsePayload structs with different object type fields.
        """
        a = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY
        )
        b = payloads.GetResponsePayload(
            object_type=enums.ObjectType.OPAQUE_DATA
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_unique_identifier(self):
        """
        Test that the inequality operator returns True when comparing two
        GetResponsePayload structs with different unique identifier fields.
        """
        a = payloads.GetResponsePayload(
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038'
        )
        b = payloads.GetResponsePayload(
            unique_identifier='49a1ca88-6bea-4fb2-ffff-7e58802c3038'
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_secrets(self):
        """
        Test that the inequality operator returns True when comparing two
        GetResponsePayload structs with different secret fields.
        """
        # TODO (peter-hamilton): Update this test case once SymmetricKeys
        # support proper field-based equality.
        a = payloads.GetResponsePayload(
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )
        b = payloads.GetResponsePayload(
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_type_mismatch(self):
        """
        Test that the inequality operators returns True when comparing two
        GetResponsePayload structs with different types.
        """
        a = payloads.GetResponsePayload()
        b = 'invalid'

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_repr(self):
        """
        Test that repr can be applied to a GetResponsePayload struct.
        """
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secrets.SymmetricKey(
                key_block=objects.KeyBlock(
                    key_format_type=misc.KeyFormatType(
                        enums.KeyFormatType.RAW
                    ),
                    key_value=objects.KeyValue(
                        key_material=objects.KeyMaterial(
                            b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                            b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                            b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                        )
                    ),
                    cryptographic_algorithm=attributes.CryptographicAlgorithm(
                        enums.CryptographicAlgorithm.TRIPLE_DES
                    ),
                    cryptographic_length=attributes.CryptographicLength(168)
                )
            )
        )

        # TODO (peter-hamilton): Update the secret portion once SymmetricKeys
        # support repr/str.
        expected = (
            "GetResponsePayload("
            "object_type=ObjectType.SYMMETRIC_KEY, "
            "unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038', "
            "secret=Struct()"
            ")"
        )
        observed = repr(payload)

        self.assertEqual(expected, observed)

    def test_str(self):
        """
        Test that str can be applied to a GetResponsePayload struct.
        """
        secret = secrets.SymmetricKey(
            key_block=objects.KeyBlock(
                key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
                key_value=objects.KeyValue(
                    key_material=objects.KeyMaterial(
                        b'\x73\x67\x57\x80\x51\x01\x2A\x6D'
                        b'\x13\x4A\x85\x5E\x25\xC8\xCD\x5E'
                        b'\x4C\xA1\x31\x45\x57\x29\xD3\xC8'
                    )
                ),
                cryptographic_algorithm=attributes.CryptographicAlgorithm(
                    enums.CryptographicAlgorithm.TRIPLE_DES
                ),
                cryptographic_length=attributes.CryptographicLength(168)
            )
        )
        payload = payloads.GetResponsePayload(
            object_type=enums.ObjectType.SYMMETRIC_KEY,
            unique_identifier='49a1ca88-6bea-4fb2-b450-7e58802c3038',
            secret=secret
        )

        # TODO (peter-hamilton): Update the secret portion once SymmetricKeys
        # support repr/str.
        expected = str({
            'object_type': enums.ObjectType.SYMMETRIC_KEY,
            'unique_identifier': '49a1ca88-6bea-4fb2-b450-7e58802c3038',
            'secret': secret
        })
        observed = str(payload)

        self.assertEqual(expected, observed)



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum as enumeration
import testtools

from kmip.core import enums
from kmip.core import exceptions
from kmip.core import primitives
from kmip.core import utils


# flake8: noqa
class DummyEnumeration(enumeration.Enum):
    SMALL     = primitives.Enumeration.MIN
    TOO_SMALL = primitives.Enumeration.MIN - 1
    LARGE     = primitives.Enumeration.MAX
    TOO_LARGE = primitives.Enumeration.MAX + 1
    INVALID   = 'invalid'


class TestEnumeration(testtools.TestCase):

    def setUp(self):
        super(TestEnumeration, self).setUp()

        self.encoding = (
            b'\x42\x00\x00\x05\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'
            b'\x00')
        self.encoding_bad_length = (
            b'\x42\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00')
        self.encoding_bad_padding = (
            b'\x42\x00\x00\x05\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'
            b'\xFF')

    def tearDown(self):
        super(TestEnumeration, self).tearDown()

    def test_init(self):
        """
        Test that an Enumeration can be instantiated.
        """
        enum = primitives.Enumeration(
            DummyEnumeration, DummyEnumeration.SMALL,
            enums.Tags.ACTIVATION_DATE)
        self.assertEqual(DummyEnumeration, enum.enum)
        self.assertEqual(DummyEnumeration.SMALL, enum.value)
        self.assertEqual(enums.Tags.ACTIVATION_DATE, enum.tag)

    def test_init_unset(self):
        """
        Test that an Enumeration can be instantiated with no input.
        """
        enum = primitives.Enumeration(DummyEnumeration)
        self.assertEqual(DummyEnumeration, enum.enum)
        self.assertEqual(None, enum.value)
        self.assertEqual(enums.Tags.DEFAULT, enum.tag)

    def test_validate_on_invalid_enum_type(self):
        """
        Test that a TypeError is thrown on input of invalid enum type
        (e.g., str).
        """
        args = ['invalid']
        kwargs = {'value': enums.Tags.DEFAULT}
        self.assertRaises(TypeError, primitives.Enumeration, *args, **kwargs)

    def test_validate_on_invalid_enum_value_type(self):
        """
        Test that a TypeError is thrown on input of invalid enum value type.
        """
        args = [DummyEnumeration]
        kwargs = {'value': enums.Tags.DEFAULT}
        self.assertRaises(TypeError, primitives.Enumeration, *args, **kwargs)

    def test_validate_on_invalid_value_type(self):
        """
        Test that a TypeError is thrown on input of invalid value type
        (e.g., str).
        """
        args = [DummyEnumeration]
        kwargs = {'value': DummyEnumeration.INVALID}
        self.assertRaises(TypeError, primitives.Enumeration, *args, **kwargs)

    def test_validate_on_invalid_value_too_big(self):
        """
        Test that a ValueError is thrown on input that is too large.
        """
        args = [DummyEnumeration]
        kwargs = {'value': DummyEnumeration.TOO_LARGE}
        self.assertRaises(ValueError, primitives.Enumeration, *args, **kwargs)

    def test_validate_on_invalid_value_too_small(self):
        """
        Test that a ValueError is thrown on input that is too small.
        """
        args = [DummyEnumeration]
        kwargs = {'value': DummyEnumeration.TOO_SMALL}
        self.assertRaises(ValueError, primitives.Enumeration, *args, **kwargs)

    def test_read(self):
        """
        Test that an Enumeration can be read from a byte stream.
        """
        stream = utils.BytearrayStream(self.encoding)
        enum = primitives.Enumeration(DummyEnumeration)
        enum.read(stream)
        self.assertEqual(DummyEnumeration.SMALL, enum.value)

    def test_read_on_invalid_length(self):
        """
        Test that an InvalidPrimitiveLength exception is thrown when attempting
        to decode an Enumeration with an invalid length.
        """
        stream = utils.BytearrayStream(self.encoding_bad_length)
        enum = primitives.Enumeration(enums.Tags)

        self.assertRaises(exceptions.InvalidPrimitiveLength, enum.read, stream)

    def test_read_on_invalid_padding(self):
        """
        Test that an InvalidPrimitiveLength exception is thrown when attempting
        to decode an Enumeration with invalid padding bytes.
        """
        stream = utils.BytearrayStream(self.encoding_bad_padding)
        enum = primitives.Enumeration(enums.Types)

        self.assertRaises(exceptions.InvalidPaddingBytes, enum.read, stream)

    def test_write(self):
        """
        Test that an Enumeration can be written to a byte stream.
        """
        stream = utils.BytearrayStream()
        enum = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        enum.write(stream)

        result = stream.read()
        self.assertEqual(len(self.encoding), len(result))
        self.assertEqual(self.encoding, result)

    def test_repr(self):
        """
        Test that the representation of an Enumeration is formatted properly.
        """
        enum = "enum={0}".format(DummyEnumeration.__name__)
        value = "value={0}".format(DummyEnumeration.SMALL)
        tag = "tag={0}".format(enums.Tags.DEFAULT)
        r = "Enumeration({0}, {1}, {2})".format(enum, value, tag)

        enum = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        self.assertEqual(r, repr(enum))

    def test_str(self):
        """
        Test that the string representation of an Enumeration is formatted
        properly.
        """
        enum = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        self.assertEqual(str(DummyEnumeration.SMALL), str(enum.value))

    def test_equal_on_equal(self):
        """
        Test that the equality operator returns True when comparing two
        Enumerations.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_equal_and_empty(self):
        """
        Test that the equality operator returns True when comparing two
        Enumerations.
        """
        a = primitives.Enumeration(DummyEnumeration)
        b = primitives.Enumeration(DummyEnumeration)

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_not_equal(self):
        """
        Test that the equality operator returns False when comparing two
        Enumerations with different values.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(DummyEnumeration, DummyEnumeration.LARGE)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_enum(self):
        """
        Test that the equality operator returns False when comparing two
        Enumerations with different enum types.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(enums.Tags, enums.Tags.DEFAULT)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_type_mismatch(self):
        """
        Test that the equality operator returns False when comparing a
        Enumeration to a non-Enumeration object.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = 'invalid'

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_not_equal_on_equal(self):
        """
        Test that the inequality operator returns False when comparing
        two Enumerations with the same values.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_equal_and_empty(self):
        """
        Test that the inequality operator returns False when comparing
        two Enumerations.
        """
        a = primitives.Enumeration(DummyEnumeration)
        b = primitives.Enumeration(DummyEnumeration)

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_not_equal(self):
        """
        Test that the inequality operator returns True when comparing two
        Enumerations with different values.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(DummyEnumeration, DummyEnumeration.LARGE)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_enum(self):
        """
        Test that the equality operator returns True when comparing two
        Enumerations with different enum types.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = primitives.Enumeration(enums.Tags, enums.Tags.DEFAULT)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_type_mismatch(self):
        """
        Test that the inequality operator returns True when comparing a
        Enumeration to a non-Enumeration object.
        """
        a = primitives.Enumeration(DummyEnumeration, DummyEnumeration.SMALL)
        b = 'invalid'

        self.assertTrue(a != b)
        self.assertTrue(b != a)



# Copyright (c) 2015 Hewlett Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from testtools import TestCase

from kmip.core import utils
from kmip.core import attributes

from kmip.core.messages import payloads


class TestActivateRequestPayload(TestCase):
    """
    Test suite for the ActivateRequestPayload class.

    Test encodings obtained from Sections 4.2 of the KMIP 1.1 Test
    Cases documentation.
    """

    def setUp(self):
        super(TestActivateRequestPayload, self).setUp()

        self.uuid = attributes.UniqueIdentifier(
            '668eff89-3010-4258-bc0e-8c402309c746')

        self.encoding_a = utils.BytearrayStream((
            b'\x42\x00\x79\x01\x00\x00\x00\x30\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x36\x36\x38\x65\x66\x66\x38\x39\x2D\x33\x30\x31\x30\x2D\x34\x32'
            b'\x35\x38\x2D\x62\x63\x30\x65\x2D\x38\x63\x34\x30\x32\x33\x30\x39'
            b'\x63\x37\x34\x36\x00\x00\x00\x00'))

    def tearDown(self):
        super(TestActivateRequestPayload, self).tearDown()

    def test_init_with_none(self):
        """
        Test that a ActivateRequestPayload object can be constructed with no
        specified value.
        """
        payloads.ActivateRequestPayload()

    def test_init_with_args(self):
        """
        Test that a ActivateRequestPayload object can be constructed with valid
        values.
        """
        payloads.ActivateRequestPayload(unique_identifier=self.uuid)

    def test_validate_with_bad_uuid_type(self):
        """
        Test that a TypeError exception is raised when an invalid UUID type
        is used to construct a ActivateRequestPayload object.
        """
        self.assertRaisesRegex(
            TypeError, "invalid unique identifier",
            payloads.ActivateRequestPayload, "not-a-uuid")

    def test_read_with_known_uuid(self):
        """
        Test that a ActivateRequestPayload object with known UUID can be read
        from a data stream.
        """
        payload = payloads.ActivateRequestPayload()
        payload.read(self.encoding_a)
        expected = '668eff89-3010-4258-bc0e-8c402309c746'
        observed = payload.unique_identifier.value

        msg = "activate UUID value mismatch"
        msg += "; expected {0}, received {1}".format(
            expected, observed)
        self.assertEqual(expected, observed, msg)

    def test_write_with_known_uuid(self):
        """
        Test that a ActivateRequestPayload object with a known UUID can be
        written to a data stream.
        """
        stream = utils.BytearrayStream()
        payload = payloads.ActivateRequestPayload(self.uuid)
        payload.write(stream)

        length_expected = len(self.encoding_a)
        length_received = len(stream)

        msg = "encoding lengths not equal"
        msg += "; expected {0}, received {1}".format(
            length_expected, length_received)
        self.assertEqual(length_expected, length_received, msg)

        msg = "encoding mismatch"
        msg += ";\nexpected:\n{0}\nreceived:\n{1}".format(self.encoding_a,
                                                          stream)

        self.assertEqual(self.encoding_a, stream, msg)


class TestActivateResponsePayload(TestCase):
    """
    Test encodings obtained from Sections 4.2 of the KMIP 1.1 Test
    Cases documentation.
    """

    def setUp(self):
        super(TestActivateResponsePayload, self).setUp()

        self.uuid = attributes.UniqueIdentifier(
            '668eff89-3010-4258-bc0e-8c402309c746')

        self.encoding_a = utils.BytearrayStream((
            b'\x42\x00\x7C\x01\x00\x00\x00\x30\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x36\x36\x38\x65\x66\x66\x38\x39\x2D\x33\x30\x31\x30\x2D\x34\x32'
            b'\x35\x38\x2D\x62\x63\x30\x65\x2D\x38\x63\x34\x30\x32\x33\x30\x39'
            b'\x63\x37\x34\x36\x00\x00\x00\x00'))

    def tearDown(self):
        super(TestActivateResponsePayload, self).tearDown()

    def test_init_with_none(self):
        """
        Test that a ActivateResponsePayload object can be constructed with no
        specified value.
        """
        payloads.ActivateResponsePayload()

    def test_init_with_args(self):
        """
        Test that a ActivateResponsePayload object can be constructed with
        valid values.
        """
        payloads.ActivateResponsePayload(unique_identifier=self.uuid)

    def test_validate_with_invalid_uuid(self):
        """
        Test that a TypeError exception is raised when an invalid Operations
        list is used to construct a ActivateResponsePayload object.
        """
        self.assertRaisesRegex(
            TypeError, "invalid unique identifier",
            payloads.ActivateResponsePayload, "not-a-uuid")

    def test_read_with_known_uuid(self):
        """
        Test that a ActivateResponsePayload object with known UUID can be read
        from a data stream.
        """
        payload = payloads.ActivateResponsePayload()
        payload.read(self.encoding_a)
        expected = '668eff89-3010-4258-bc0e-8c402309c746'
        observed = payload.unique_identifier.value

        msg = "activate UUID value mismatch"
        msg += "; expected {0}, received {1}".format(
            expected, observed)
        self.assertEqual(expected, observed, msg)

    def test_write_with_known_uuid(self):
        """
        Test that a ActivateResponsePayload object with a known UUID can be
        written to a data stream.
        """
        stream = utils.BytearrayStream()
        payload = payloads.ActivateResponsePayload(self.uuid)
        payload.write(stream)

        length_expected = len(self.encoding_a)
        length_received = len(stream)

        msg = "encoding lengths not equal"
        msg += "; expected {0}, received {1}".format(
            length_expected, length_received)
        self.assertEqual(length_expected, length_received, msg)

        msg = "encoding mismatch"
        msg += ";\nexpected:\n{0}\nreceived:\n{1}".format(self.encoding_a,
                                                          stream)

        self.assertEqual(self.encoding_a, stream, msg)



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kmip.core import enums


class KmipError(Exception):
    """
    A generic KMIP error that is the base for the KMIP error hierarchy.
    """

    def __init__(self,
                 status=enums.ResultStatus.OPERATION_FAILED,
                 reason=enums.ResultReason.GENERAL_FAILURE,
                 message='A general failure occurred.'):
        """
        Create a KmipError exception.

        Args:
            status (ResultStatus): An enumeration detailing the result outcome.
            reason (ResultReason): An enumeration giving the status rationale.
            message (string): A string containing more information about the
                error.
        """
        super(KmipError, self).__init__(message)
        self.status = status
        self.reason = reason

    def __eq__(self, other):
        if isinstance(other, KmipError):
            if str(self) != str(other):
                return False
            elif self.status != other.status:
                return False
            elif self.reason != other.reason:
                return False
            else:
                return True
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, KmipError):
            return self == other
        else:
            return NotImplemented


class CryptographicFailure(KmipError):
    """
    An error generated when problems occur with cryptographic operations.
    """

    def __init__(self, message):
        """
        Create a CryptographicFailure exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(CryptographicFailure, self).__init__(
            reason=enums.ResultReason.CRYPTOGRAPHIC_FAILURE,
            message=message
        )


class EncodingOptionError(KmipError):
    """
    An encoding error generated during key wrapping.

    This error is generated during key wrapping when a requested encoding
    option is not supported or is incompatible with other settings in the
    key wrapping request (e.g., attributes are requested with the key but
    the encoding does not support wrapping attributes with the key value).
    """
    def __init__(self, message):
        """
        Create an EncodingOptionError.

        Args:
            message (string): A string containing information about the error.
        """
        super(EncodingOptionError, self).__init__(
            reason=enums.ResultReason.ENCODING_OPTION_ERROR,
            message=message
        )


class IllegalOperation(KmipError):
    """
    An error generated when an improper operation is attempted. The operation
    can be 'illegal' for various reasons, including invalid permissions or
    literal object/operation mismatch (e.g., a Template item cannot be
    activated with the Activate operation since it has no state).
    """
    def __init__(self, message):
        """
        Create an IllegalOperation exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(IllegalOperation, self).__init__(
            reason=enums.ResultReason.ILLEGAL_OPERATION,
            message=message
        )


class IndexOutOfBounds(KmipError):
    """
    An error generated when exceeding the attribute instance limit.
    """

    def __init__(self, message):
        """
        Create an IndexOutOfBounds exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(IndexOutOfBounds, self).__init__(
            reason=enums.ResultReason.INDEX_OUT_OF_BOUNDS,
            message=message
        )


class InvalidField(KmipError):
    """
    An error generated when an invalid field value is processed.
    """

    def __init__(self, message):
        """
        Create an InvalidField exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(InvalidField, self).__init__(
            reason=enums.ResultReason.INVALID_FIELD,
            message=message
        )


class InvalidMessage(KmipError):
    """
    An error generated when an invalid message is processed.
    """

    def __init__(self, message):
        """
        Create an InvalidMessage exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(InvalidMessage, self).__init__(
            reason=enums.ResultReason.INVALID_MESSAGE,
            message=message
        )


class ItemNotFound(KmipError):
    """
    An error generated when a request item cannot be located.
    """

    def __init__(self, message):
        """
        Create an ItemNotFound exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(ItemNotFound, self).__init__(
            reason=enums.ResultReason.ITEM_NOT_FOUND,
            message=message
        )


class KeyCompressionTypeNotSupported(KmipError):
    """
    An error generated when dealing with unsupported key compression types
    and operations.
    """

    def __init__(self, message):
        """
        Create a KeyCompressionTypeNotSupported exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(KeyCompressionTypeNotSupported, self).__init__(
            reason=enums.ResultReason.KEY_COMPRESSION_TYPE_NOT_SUPPORTED,
            message=message
        )


class KeyFormatTypeNotSupported(KmipError):
    """
    An error generated when dealing with unsupported key formats
    and operations.
    """

    def __init__(self, message):
        """
        Create a KeyFormatTypeNotSupported exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(KeyFormatTypeNotSupported, self).__init__(
            reason=enums.ResultReason.KEY_FORMAT_TYPE_NOT_SUPPORTED,
            message=message
        )


class OperationFailure(KmipError):
    """
    An exception raised upon the failure of a KMIP appliance operation.
    """
    def __init__(self, status, reason, message):
        """
        Construct the error message and attributes for the KMIP operation
        failure.

        Args:
            status: a ResultStatus enumeration
            reason: a ResultReason enumeration
            message: a string providing additional error information
        """
        super(OperationFailure, self).__init__(status, reason, message)


class OperationNotSupported(KmipError):
    """
    An error generated when an unsupported operation is invoked.
    """

    def __init__(self, message):
        """
        Create an OperationNotSupported exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(OperationNotSupported, self).__init__(
            reason=enums.ResultReason.OPERATION_NOT_SUPPORTED,
            message=message
        )


class PermissionDenied(KmipError):
    """
    An error generated when permission constraints are violated.
    """

    def __init__(self, message):
        """
        Create a PermissionDenied exception.

        Args:
            message (string): A string containing information about the error.
        """
        super(PermissionDenied, self).__init__(
            reason=enums.ResultReason.PERMISSION_DENIED,
            message=message
        )


class AttributeNotSupported(Exception):
    """
    An error generated when an unsupported attribute is processed.
    """
    pass


class ConfigurationError(Exception):
    """
    An error generated when a problem occurs with a client or server
    configuration.
    """
    pass


class ConnectionClosed(Exception):
    """
    An exception generated when attempting to use a connection that has been
    closed.
    """
    pass


class NetworkingError(Exception):
    """
    An error generated when a problem occurs with client or server networking
    activity.
    """
    pass


class InvalidKmipEncoding(Exception):
    """
    An exception raised when processing invalid KMIP message encodings.
    """
    pass


class InvalidPaddingBytes(Exception):
    """
    An exception raised for errors when processing the padding bytes of
    primitive encodings.
    """
    pass


class InvalidPrimitiveLength(Exception):
    """
    An exception raised for errors when processing primitives with invalid
    lengths.
    """
    pass


class ShutdownError(Exception):
    """
    An error generated when a problem occurs with shutting down the server.
    """


class VersionNotSupported(Exception):
    """
    An error generated when an unsupported KMIP version is referenced.
    """


class StreamNotEmptyError(Exception):
    def __init__(self, cls, extra):
        super(StreamNotEmptyError, self).__init__()

        self.cls = cls
        self.extra = extra

    def __str__(self):
        msg = "Invalid length used to read {0}, bytes remaining: {1}"
        return msg.format(self.cls, self.extra)


class ReadValueError(Exception):
    def __init__(self, cls, attr, exp, recv):
        super(ReadValueError, self).__init__()

        self.cls = cls
        self.attr = attr
        self.exp = exp
        self.recv = recv

    def __str__(self):
        msg = "Tried to read {0}.{1}: expected {2}, received {3}"
        return msg.format(self.cls, self.attr, self.exp, self.recv)


class WriteOverflowError(Exception):
    def __init__(self, cls, attr, exp, recv):
        super(WriteOverflowError, self).__init__()

        self.cls = cls
        self.attr = attr
        self.exp = exp
        self.recv = recv

    def __str__(self):
        msg = "Tried to write {0}.{1} with too many bytes: "
        msg += "expected {2}, received {3}"
        return msg.format(self.cls, self.attr, self.exp, self.recv)


class KMIPServerZombieError(Exception):
    """KMIP server error for hung and persistent live KMIP servers."""
    def __init__(self, pid):
        super(KMIPServerZombieError, self).__init__()

        self.message = 'KMIP server alive after termination: PID {0}'.format(
            pid
        )

    def __str__(self):
        return self.message


class KMIPServerSuicideError(Exception):
    """KMIP server error for prematurely dead KMIP servers."""
    def __init__(self, pid):
        super(KMIPServerSuicideError, self).__init__()

        self.message = 'KMIP server dead prematurely: PID {0}'.format(pid)

    def __str__(self):
        return self.message


class ErrorStrings:
    BAD_EXP_RECV = "Bad {0} {1}: expected {2}, received {3}"
    BAD_ENCODING = "Bad {0} {1}: encoding mismatch"



# Copyright (c) 2019 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import testtools

from kmip.core import enums
from kmip.core import exceptions
from kmip.core import misc
from kmip.core import objects
from kmip.core import primitives
from kmip.core import secrets
from kmip.core import utils


class TestSplitKey(testtools.TestCase):
    """
    Test suite for the SplitKey secret object.
    """

    def setUp(self):
        super(TestSplitKey, self).setUp()

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite. The Prime Field Size was manually added.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        #     Key Part Identifier - 1
        #     Split Key Threshold - 2
        #     Split Key Method - Polynomial Sharing GF 2^8
        #     Prime Field Size - 104729
        #     Key Block
        #         Key Format Type - Raw
        #         Key Value
        #             Key Material - 0x66C46A7754F94DE420C7B1A7FFF5EC56
        #         Cryptographic Algorithm - AES
        #         Cryptographic Length - 128
        self.full_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\xA8'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8C\x02\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x8A\x05\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x62\x04\x00\x00\x00\x08\x00\x00\x00\x00\x00\x01\x99\x19'
            b'\x42\x00\x40\x01\x00\x00\x00\x50'
            b'\x42\x00\x42\x05\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x45\x01\x00\x00\x00\x18'
            b'\x42\x00\x43\x08\x00\x00\x00\x10'
            b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
            b'\x42\x00\x28\x05\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\x2A\x02\x00\x00\x00\x04\x00\x00\x00\x80\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Key Part Identifier - 1
        self.no_split_key_parts_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x10'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        self.no_key_part_identifier_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x10'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        #     Key Part Identifier - 1
        self.no_split_key_threshold_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x20'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        #     Key Part Identifier - 1
        #     Split Key Threshold - 2
        self.no_split_key_method_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x30'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8C\x02\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        #     Key Part Identifier - 1
        #     Split Key Threshold - 2
        #     Split Key Method - Polynomial Sharing Prime Field
        self.no_prime_field_size_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x40'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8C\x02\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x8A\x05\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
        )

        # This encoding was adapted from test case TC-SJ-2-20 from the KMIP
        # 2.0 test suite.
        #
        # This encoding matches the following set of values:
        # SplitKey
        #     Split Key Parts - 4
        #     Key Part Identifier - 1
        #     Split Key Threshold - 2
        #     Split Key Method - Polynomial Sharing GF 2^8
        self.no_key_block_encoding = utils.BytearrayStream(
            b'\x42\x00\x89\x01\x00\x00\x00\x40'
            b'\x42\x00\x8B\x02\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
            b'\x42\x00\x44\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8C\x02\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x8A\x05\x00\x00\x00\x04\x00\x00\x00\x04\x00\x00\x00\x00'
        )

    def tearDown(self):
        super(TestSplitKey, self).tearDown()

    def test_invalid_split_key_parts(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the split key parts of a SplitKey object.
        """
        kwargs = {"split_key_parts": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The split key parts must be an integer.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "split_key_parts",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The split key parts must be an integer.",
            setattr,
            *args
        )

    def test_invalid_key_part_identifier(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the key part identifier of a SplitKey object.
        """
        kwargs = {"key_part_identifier": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The key part identifier must be an integer.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "key_part_identifier",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The key part identifier must be an integer.",
            setattr,
            *args
        )

    def test_invalid_split_key_threshold(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the split key threshold of a SplitKey object.
        """
        kwargs = {"split_key_threshold": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The split key threshold must be an integer.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "split_key_threshold",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The split key threshold must be an integer.",
            setattr,
            *args
        )

    def test_invalid_split_key_method(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the split key method of a SplitKey object.
        """
        kwargs = {"split_key_method": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The split key method must be a SplitKeyMethod enumeration.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "split_key_method",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The split key method must be a SplitKeyMethod enumeration.",
            setattr,
            *args
        )

    def test_invalid_prime_field_size(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the prime field size of a SplitKey object.
        """
        kwargs = {"prime_field_size": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The prime field size must be an integer.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "prime_field_size",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The prime field size must be an integer.",
            setattr,
            *args
        )

    def test_invalid_key_block(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the key block of a SplitKey object.
        """
        kwargs = {"key_block": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "The key block must be a KeyBlock structure.",
            secrets.SplitKey,
            **kwargs
        )

        args = (
            secrets.SplitKey(),
            "key_block",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "The key block must be a KeyBlock structure.",
            setattr,
            *args
        )

    def test_read(self):
        """
        Test that a SplitKey object can be read from a buffer.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.split_key_parts)
        self.assertIsNone(split_key.key_part_identifier)
        self.assertIsNone(split_key.split_key_threshold)
        self.assertIsNone(split_key.split_key_method)
        self.assertIsNone(split_key.prime_field_size)
        self.assertIsNone(split_key.key_block)

        split_key.read(self.full_encoding)

        self.assertEqual(4, split_key.split_key_parts)
        self.assertEqual(1, split_key.key_part_identifier)
        self.assertEqual(2, split_key.split_key_threshold)
        self.assertEqual(
            enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            split_key.split_key_method
        )
        self.assertEqual(104729, split_key.prime_field_size)
        self.assertIsInstance(split_key.key_block, objects.KeyBlock)
        self.assertEqual(
            enums.KeyFormatType.RAW,
            split_key.key_block.key_format_type.value
        )
        self.assertIsInstance(split_key.key_block.key_value, objects.KeyValue)
        self.assertIsInstance(
            split_key.key_block.key_value.key_material,
            primitives.ByteString
        )
        self.assertEqual(
            (
                b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
            ),
            split_key.key_block.key_value.key_material.value
        )
        self.assertEqual(
            enums.CryptographicAlgorithm.AES,
            split_key.key_block.cryptographic_algorithm.value
        )
        self.assertEqual(128, split_key.key_block.cryptographic_length.value)

    def test_read_missing_split_key_parts(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the split key parts are missing from the
        encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.split_key_parts)

        args = (self.no_split_key_parts_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the SplitKeyParts field.",
            split_key.read,
            *args
        )

    def test_read_missing_key_part_identifier(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the key part identifier is missing from the
        encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.key_part_identifier)

        args = (self.no_key_part_identifier_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the KeyPartIdentifier field.",
            split_key.read,
            *args
        )

    def test_read_missing_split_key_threshold(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the split key threshold is missing from the
        encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.split_key_threshold)

        args = (self.no_split_key_threshold_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the SplitKeyThreshold field.",
            split_key.read,
            *args
        )

    def test_read_missing_split_key_method(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the split key method is missing from the
        encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.split_key_method)

        args = (self.no_split_key_method_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the SplitKeyMethod field.",
            split_key.read,
            *args
        )

    def test_read_missing_prime_field_size(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the prime field size is missing from the
        encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.prime_field_size)

        args = (self.no_prime_field_size_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the PrimeFieldSize "
            "field. This field is required when the SplitKeyMethod is "
            "PolynomialSharingPrimeField.",
            split_key.read,
            *args
        )

    def test_read_missing_key_block(self):
        """
        Test that an InvalidKmipEncoding error is raised during the decoding
        of a SplitKey object when the key block is missing from the encoding.
        """
        split_key = secrets.SplitKey()

        self.assertIsNone(split_key.key_block)

        args = (self.no_key_block_encoding, )
        self.assertRaisesRegex(
            exceptions.InvalidKmipEncoding,
            "The SplitKey encoding is missing the KeyBlock field.",
            split_key.read,
            *args
        )

    def test_write(self):
        """
        Test that a SplitKey object can be written to a buffer.
        """
        # TODO (peter-hamilton) Update this test when the KeyBlock supports
        # generic key format type and key value/material values.
        key_block = objects.KeyBlock(
            key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
            key_value=objects.KeyValue(
                key_material=objects.KeyMaterial(
                    value=(
                        b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                        b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
                    )
                )
            ),
            cryptographic_algorithm=primitives.Enumeration(
                enums.CryptographicAlgorithm,
                value=enums.CryptographicAlgorithm.AES,
                tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
            ),
            cryptographic_length=primitives.Integer(
                value=128,
                tag=enums.Tags.CRYPTOGRAPHIC_LENGTH
            )
        )
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            prime_field_size=104729,
            key_block=key_block
        )

        stream = utils.BytearrayStream()
        split_key.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_missing_split_key_parts(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the split key parts field.
        """
        split_key = secrets.SplitKey(key_part_identifier=1)

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the SplitKeyParts field.",
            split_key.write,
            *args
        )

    def test_write_missing_key_part_identifier(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the key part identifier
        field.
        """
        split_key = secrets.SplitKey(split_key_parts=4)

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the KeyPartIdentifier field.",
            split_key.write,
            *args
        )

    def test_write_missing_split_key_threshold(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the split key threshold
        field.
        """
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1
        )

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the SplitKeyThreshold field.",
            split_key.write,
            *args
        )

    def test_write_missing_split_key_method(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the split key method field.
        """
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2
        )

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the SplitKeyMethod field.",
            split_key.write,
            *args
        )

    def test_write_missing_prime_field_size(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the prime field size field.
        """
        split_key_method = enums.SplitKeyMethod.POLYNOMIAL_SHARING_PRIME_FIELD
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=split_key_method
        )

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the PrimeFieldSize field. "
            "This field is required when the SplitKeyMethod is "
            "PolynomialSharingPrimeField.",
            split_key.write,
            *args
        )

    def test_write_missing_key_block(self):
        """
        Test that an InvalidField error is raised during the encoding of a
        SplitKey object when the object is missing the key block field.
        """
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8
        )

        stream = utils.BytearrayStream()
        args = (stream, )
        self.assertRaisesRegex(
            exceptions.InvalidField,
            "The SplitKey object is missing the KeyBlock field.",
            split_key.write,
            *args
        )

    def test_repr(self):
        """
        Test that repr can be applied to a SplitKey object.
        """
        key_block = objects.KeyBlock(
            key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
            key_value=objects.KeyValue(
                key_material=objects.KeyMaterial(
                    value=(
                        b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                        b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
                    )
                )
            ),
            cryptographic_algorithm=primitives.Enumeration(
                enums.CryptographicAlgorithm,
                value=enums.CryptographicAlgorithm.AES,
                tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
            ),
            cryptographic_length=primitives.Integer(
                value=128,
                tag=enums.Tags.CRYPTOGRAPHIC_LENGTH
            )
        )
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            prime_field_size=104729,
            key_block=key_block
        )

        args = [
            "split_key_parts=4",
            "key_part_identifier=1",
            "split_key_threshold=2",
            "split_key_method=SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8",
            "prime_field_size=104729",
            "key_block=Struct()"
        ]
        self.assertEqual(
            "SplitKey({})".format(", ".join(args)),
            repr(split_key)
        )

    def test_str(self):
        """
        Test that str can be applied to a SplitKey object.
        """
        key_block = objects.KeyBlock(
            key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
            key_value=objects.KeyValue(
                key_material=objects.KeyMaterial(
                    value=(
                        b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                        b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
                    )
                )
            ),
            cryptographic_algorithm=primitives.Enumeration(
                enums.CryptographicAlgorithm,
                value=enums.CryptographicAlgorithm.AES,
                tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
            ),
            cryptographic_length=primitives.Integer(
                value=128,
                tag=enums.Tags.CRYPTOGRAPHIC_LENGTH
            )
        )
        split_key = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            prime_field_size=104729,
            key_block=key_block
        )

        args = [
            ("split_key_parts", 4),
            ("key_part_identifier", 1),
            ("split_key_threshold", 2),
            (
                "split_key_method",
                enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8
            ),
            ("prime_field_size", 104729),
            ("key_block", str(key_block))
        ]
        value = "{}".format(
            ", ".join(['"{}": {}'.format(arg[0], arg[1]) for arg in args])
        )
        self.assertEqual(
            "{" + value + "}",
            str(split_key)
        )

    def test_comparison(self):
        """
        Test that the equality/inequality operators return True/False when
        comparing two SplitKey objects with the same data.
        """
        a = secrets.SplitKey()
        b = secrets.SplitKey()

        self.assertTrue(a == b)
        self.assertTrue(b == a)
        self.assertFalse(a != b)
        self.assertFalse(b != a)

        a = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            prime_field_size=104729,
            key_block=objects.KeyBlock(
                key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
                key_value=objects.KeyValue(
                    key_material=objects.KeyMaterial(
                        value=(
                            b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                            b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
                        )
                    )
                ),
                cryptographic_algorithm=primitives.Enumeration(
                    enums.CryptographicAlgorithm,
                    value=enums.CryptographicAlgorithm.AES,
                    tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
                ),
                cryptographic_length=primitives.Integer(
                    value=128,
                    tag=enums.Tags.CRYPTOGRAPHIC_LENGTH
                )
            )
        )
        b = secrets.SplitKey(
            split_key_parts=4,
            key_part_identifier=1,
            split_key_threshold=2,
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8,
            prime_field_size=104729,
            key_block=objects.KeyBlock(
                key_format_type=misc.KeyFormatType(enums.KeyFormatType.RAW),
                key_value=objects.KeyValue(
                    key_material=objects.KeyMaterial(
                        value=(
                            b'\x66\xC4\x6A\x77\x54\xF9\x4D\xE4'
                            b'\x20\xC7\xB1\xA7\xFF\xF5\xEC\x56'
                        )
                    )
                ),
                cryptographic_algorithm=primitives.Enumeration(
                    enums.CryptographicAlgorithm,
                    value=enums.CryptographicAlgorithm.AES,
                    tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
                ),
                cryptographic_length=primitives.Integer(
                    value=128,
                    tag=enums.Tags.CRYPTOGRAPHIC_LENGTH
                )
            )
        )

        self.assertTrue(a == b)
        self.assertTrue(b == a)
        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_comparison_on_different_split_key_parts(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different split key parts.
        """
        a = secrets.SplitKey(split_key_parts=4)
        b = secrets.SplitKey(split_key_parts=6)

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_comparison_on_different_key_part_identifiers(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different key part identifiers.
        """
        a = secrets.SplitKey(key_part_identifier=1)
        b = secrets.SplitKey(key_part_identifier=2)

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_comparison_on_different_split_key_thresholds(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different split key thresholds.
        """
        a = secrets.SplitKey(split_key_threshold=3)
        b = secrets.SplitKey(split_key_threshold=4)

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_comparison_on_different_split_key_methods(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different split key methods.
        """
        a = secrets.SplitKey(split_key_method=enums.SplitKeyMethod.XOR)
        b = secrets.SplitKey(
            split_key_method=enums.SplitKeyMethod.POLYNOMIAL_SHARING_GF_2_8
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_comparison_on_different_prime_field_sizes(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different prime field sizes.
        """
        a = secrets.SplitKey(prime_field_size=104723)
        b = secrets.SplitKey(prime_field_size=104729)

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    # TODO (peter-hamilton) Fill in this test once the KeyBlock supports the
    # comparison operators.
    def test_comparison_on_different_key_blocks(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different key blocks.
        """
        self.skipTest(
            "The KeyBlock structure does not support the comparison operators."
        )

    def test_comparison_on_type_mismatch(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two SplitKey objects with different types.
        """
        a = secrets.SplitKey()
        b = "invalid"

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)



# Copyright (c) 2016 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import binascii
import logging
import socket
import struct
import threading
import time

from cryptography import x509

from kmip.core import enums
from kmip.core import exceptions
from kmip.core.messages import contents
from kmip.core.messages import messages
from kmip.core import utils

from kmip.services.server import auth


class KmipSession(threading.Thread):
    """
    A session thread representing a single KMIP client/server interaction.
    """

    def __init__(self,
                 engine,
                 connection,
                 address,
                 name=None,
                 enable_tls_client_auth=True,
                 auth_settings=None):
        """
        Create a KmipSession.

        Args:
            engine (KmipEngine): A reference to the central server application
                that handles message processing. Required.
            connection (socket): A client socket.socket TLS connection
                representing a new KMIP connection. Required.
            address (tuple): The address tuple produced with the session
                connection. Contains the IP address and port number of the
                remote connection endpoint. Required.
            name (str): The name of the KmipSession. Optional, defaults to
                None.
            enable_tls_client_auth (bool): A flag that enables a strict check
                for the client auth flag in the extended key usage extension
                in client certificates when establishing the client/server TLS
                connection. Optional, defaults to True.
            auth_settings (list): A list of tuples, each containing (1) the
                name of the 'auth:' settings block from the server config file,
                and (2) a dictionary of configuration settings for a specific
                authentication plugin. Optional, defaults to None.
        """
        super(KmipSession, self).__init__(
            group=None,
            target=None,
            name=name,
            args=(),
            kwargs={}
        )

        self._logger = logging.getLogger(
            'kmip.server.session.{0}'.format(self.name)
        )

        self._engine = engine
        self._connection = connection
        self._address = address

        self._enable_tls_client_auth = enable_tls_client_auth
        self._auth_settings = [] if auth_settings is None else auth_settings

        self._session_time = time.time()
        self._max_buffer_size = 4096
        self._max_request_size = 1048576
        self._max_response_size = 1048576

    def run(self):
        """
        The main thread routine executed by invoking thread.start.

        This method manages the new client connection, running a message
        handling loop. Once this method completes, the thread is finished.
        """
        self._logger.info("Starting session: {0}".format(self.name))

        try:
            self._connection.do_handshake()
        except Exception as e:
            self._logger.info("Failure running TLS handshake")
            self._logger.exception(e)
        else:
            while True:
                try:
                    self._handle_message_loop()
                except exceptions.ConnectionClosed as e:
                    break
                except Exception as e:
                    self._logger.info("Failure handling message loop")
                    self._logger.exception(e)

        self._connection.shutdown(socket.SHUT_RDWR)
        self._connection.close()
        self._logger.info("Stopping session: {0}".format(self.name))

    def _handle_message_loop(self):
        request_data = self._receive_request()
        request = messages.RequestMessage()

        max_size = self._max_response_size
        kmip_version = contents.protocol_version_to_kmip_version(
            self._engine.default_protocol_version
        )

        try:
            if hasattr(self._connection, 'shared_ciphers'):
                shared_ciphers = self._connection.shared_ciphers()
                self._logger.debug(
                    "Possible session ciphers: {0}".format(len(shared_ciphers))
                )
                for cipher in shared_ciphers:
                    self._logger.debug(cipher)
            self._logger.debug(
                "Session cipher selected: {0}".format(
                    self._connection.cipher()
                )
            )

            certificate = auth.get_certificate_from_connection(
                self._connection
            )
            if certificate is None:
                raise exceptions.PermissionDenied(
                    "The client certificate could not be loaded from the "
                    "session connection."
                )

            if self._enable_tls_client_auth:
                extension = auth.get_extended_key_usage_from_certificate(
                    certificate
                )
                if extension is None:
                    raise exceptions.PermissionDenied(
                        "The extended key usage extension is missing from "
                        "the client certificate."
                    )
                if x509.oid.ExtendedKeyUsageOID.CLIENT_AUTH not in extension:
                    raise exceptions.PermissionDenied(
                        "The extended key usage extension is not marked for "
                        "client authentication in the client certificate."
                    )

            request.read(request_data, kmip_version=kmip_version)
        except exceptions.PermissionDenied as e:
            self._logger.warning("Failure verifying the client certificate.")
            self._logger.exception(e)
            response = self._engine.build_error_response(
                contents.ProtocolVersion(1, 0),
                enums.ResultReason.AUTHENTICATION_NOT_SUCCESSFUL,
                "Error verifying the client certificate. "
                "See server logs for more information."
            )
        except Exception as e:
            self._logger.warning("Failure parsing request message.")
            self._logger.exception(e)
            response = self._engine.build_error_response(
                contents.ProtocolVersion(1, 0),
                enums.ResultReason.INVALID_MESSAGE,
                "Error parsing request message. See server logs for more "
                "information."
            )
        else:
            try:
                client_identity = self.authenticate(certificate, request)
                self._logger.info(
                    "Session client identity: {}".format(client_identity[0])
                )
            except Exception:
                self._logger.warning("Authentication failed.")
                response = self._engine.build_error_response(
                    request.request_header.protocol_version,
                    enums.ResultReason.AUTHENTICATION_NOT_SUCCESSFUL,
                    "An error occurred during client authentication. "
                    "See server logs for more information."
                )
            else:
                try:
                    results = self._engine.process_request(
                        request,
                        client_identity
                    )
                    response, max_response_size, protocol_version = results
                    kmip_version = contents.protocol_version_to_kmip_version(
                        protocol_version
                    )

                    if max_response_size:
                        max_size = max_response_size
                except exceptions.KmipError as e:
                    response = self._engine.build_error_response(
                        request.request_header.protocol_version,
                        e.reason,
                        str(e)
                    )
                except Exception as e:
                    self._logger.warning(
                        "An unexpected error occurred while processing "
                        "request."
                    )
                    self._logger.exception(e)
                    response = self._engine.build_error_response(
                        request.request_header.protocol_version,
                        enums.ResultReason.GENERAL_FAILURE,
                        "An unexpected error occurred while processing "
                        "request. See server logs for more information."
                    )

        response_data = utils.BytearrayStream()
        response.write(response_data, kmip_version=kmip_version)

        if len(response_data) > max_size:
            self._logger.warning(
                "Response message length too large: "
                "{0} bytes, max {1} bytes".format(
                    len(response_data),
                    self._max_response_size
                )
            )
            response = self._engine.build_error_response(
                request.request_header.protocol_version,
                enums.ResultReason.RESPONSE_TOO_LARGE,
                "Response message length too large. See server logs for "
                "more information."
            )
            response_data = utils.BytearrayStream()
            response.write(response_data, kmip_version=kmip_version)

        self._send_response(response_data.buffer)

    def authenticate(self, certificate, request):
        credentials = []
        if request.request_header.authentication is not None:
            credentials = request.request_header.authentication.credentials

        plugin_enabled = False

        for auth_settings in self._auth_settings:
            plugin_name, plugin_config = auth_settings

            if plugin_name.startswith("auth:slugs"):
                if plugin_config.get("enabled") == "True":
                    plugin_enabled = True
                    plugin = auth.SLUGSConnector(plugin_config.get("url"))
                    self._logger.debug(
                        "Authenticating with plugin: {}".format(plugin_name)
                    )
                    try:
                        client_identity = plugin.authenticate(
                            certificate,
                            (self._address, self._session_time),
                            credentials
                        )
                    except Exception as e:
                        self._logger.warning(
                            "Authentication failed."
                        )
                        self._logger.error(e)
                        self._logger.exception(e)
                    else:
                        self._logger.debug(
                            "Authentication succeeded for client identity: "
                            "{}".format(client_identity[0])
                        )
                        return client_identity
            else:
                self._logger.warning(
                    "Authentication plugin '{}' is not "
                    "supported.".format(plugin_name)
                )

        if not plugin_enabled:
            self._logger.debug(
                "No authentication plugins are enabled. The client identity "
                "will be extracted from the client certificate."
            )
            try:
                client_identity = auth.get_client_identity_from_certificate(
                    certificate
                )
            except Exception as e:
                self._logger.warning("Client identity extraction failed.")
                self._logger.exception(e)
            else:
                self._logger.debug(
                    "Extraction succeeded for client identity: {}".format(
                        client_identity
                    )
                )
                return tuple([client_identity, None])

        raise exceptions.PermissionDenied("Authentication failed.")

    def _receive_request(self):
        header = self._receive_bytes(8)
        message_size = struct.unpack('!I', header[4:])[0]

        payload = self._receive_bytes(message_size)
        data = utils.BytearrayStream(header + payload)

        return data

    def _receive_bytes(self, message_size):
        bytes_received = 0
        message = b''

        while bytes_received < message_size:
            partial_message = self._connection.recv(
                min(message_size - bytes_received, self._max_buffer_size)
            )

            if partial_message is None:
                break
            elif len(partial_message) == 0:
                raise exceptions.ConnectionClosed()
            else:
                bytes_received += len(partial_message)
                message += partial_message

        if bytes_received != message_size:
            raise ValueError(
                "Invalid KMIP message received. Actual message length "
                "does not match the advertised header length."
            )
        else:
            self._logger.debug(
                "Request encoding: {}".format(binascii.hexlify(message))
            )
            return message

    def _send_response(self, data):
        if len(data) > 0:
            self._logger.debug(
                "Response encoding: {}".format(binascii.hexlify(bytes(data)))
            )
            self._connection.sendall(bytes(data))



# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import sys
import warnings

from kmip.core import enums
from kmip.pie import client
from kmip.pie import objects
from kmip.pie.client import ProxyKmipClient as KmipClient

# Dynamically set __version__
version_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "version.py"
)
with open(version_path, 'r') as f:
    m = re.search(
        r"^__version__ = \"(\d+\.\d+\..*)\"$",
        f.read(),
        re.MULTILINE
    )
    __version__ = m.group(1)


__all__ = [
    'client',
    'core',
    'demos',
    'enums',
    'KmipClient',
    'objects',
    'services'
]


if sys.version_info[:2] == (2, 7):
    warnings.warn(
        (
            "PyKMIP will drop support for Python 2.7 in a future release. "
            "Please upgrade to a newer version of Python (3.5+ preferred)."
        ),
        PendingDeprecationWarning
    )

if sys.version_info[:2] == (3, 4):
    warnings.warn(
        (
            "PyKMIP will drop support for Python 3.4 in a future release. "
            "Please upgrade to a newer version of Python (3.5+ preferred)."
        ),
        PendingDeprecationWarning
    )



# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from six.moves import xrange

from testtools import TestCase

from kmip.core import utils

from kmip.core.messages.contents import ProtocolVersion
from kmip.core.messages import payloads


class TestDiscoverVersionsRequestPayload(TestCase):

    def setUp(self):
        super(TestDiscoverVersionsRequestPayload, self).setUp()

        self.protocol_versions_empty = list()
        self.protocol_versions_one = list()
        self.protocol_versions_one.append(ProtocolVersion(1, 0))
        self.protocol_versions_two = list()
        self.protocol_versions_two.append(ProtocolVersion(1, 1))
        self.protocol_versions_two.append(ProtocolVersion(1, 0))

        self.encoding_empty = utils.BytearrayStream((
            b'\x42\x00\x79\x01\x00\x00\x00\x00'))
        self.encoding_one = utils.BytearrayStream((
            b'\x42\x00\x79\x01\x00\x00\x00\x28\x42\x00\x69\x01\x00\x00\x00\x20'
            b'\x42\x00\x6A\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x6B\x02\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'
            b'\x00'))
        self.encoding_two = utils.BytearrayStream((
            b'\x42\x00\x79\x01\x00\x00\x00\x50\x42\x00\x69\x01\x00\x00\x00\x20'
            b'\x42\x00\x6A\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x6B\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x69\x01\x00\x00\x00\x20\x42\x00\x6A\x02\x00\x00\x00\x04'
            b'\x00\x00\x00\x01\x00\x00\x00\x00\x42\x00\x6B\x02\x00\x00\x00\x04'
            b'\x00\x00\x00\x00\x00\x00\x00\x00'))

    def tearDown(self):
        super(TestDiscoverVersionsRequestPayload, self).tearDown()

    def test_init_with_none(self):
        payloads.DiscoverVersionsRequestPayload()

    def test_init_with_args(self):
        payloads.DiscoverVersionsRequestPayload(
            self.protocol_versions_empty)

    def test_validate_with_invalid_protocol_versions(self):
        kwargs = {'protocol_versions': 'invalid'}
        self.assertRaisesRegex(
            TypeError, "invalid protocol versions list",
            payloads.DiscoverVersionsRequestPayload, **kwargs)

    def test_validate_with_invalid_protocol_version(self):
        kwargs = {'protocol_versions': ['invalid']}
        self.assertRaisesRegex(
            TypeError, "invalid protocol version",
            payloads.DiscoverVersionsRequestPayload, **kwargs)

    def _test_read(self, stream, payload, protocol_versions):
        payload.read(stream)
        expected = len(protocol_versions)
        observed = len(payload.protocol_versions)

        msg = "protocol versions list decoding mismatch"
        msg += "; expected {0} results, received {1}".format(
            expected, observed)
        self.assertEqual(expected, observed, msg)

        for i in xrange(len(protocol_versions)):
            expected = protocol_versions[i]
            observed = payload.protocol_versions[i]

            msg = "protocol version decoding mismatch"
            msg += "; expected {0}, received {1}".format(expected, observed)
            self.assertEqual(expected, observed, msg)

    def test_read_with_empty_protocol_list(self):
        stream = self.encoding_empty
        payload = payloads.DiscoverVersionsRequestPayload()
        protocol_versions = self.protocol_versions_empty

        self._test_read(stream, payload, protocol_versions)

    def test_read_with_one_protocol_version(self):
        stream = self.encoding_one
        payload = payloads.DiscoverVersionsRequestPayload()
        protocol_versions = self.protocol_versions_one

        self._test_read(stream, payload, protocol_versions)

    def test_read_with_two_protocol_versions(self):
        stream = self.encoding_two
        payload = payloads.DiscoverVersionsRequestPayload()
        protocol_versions = self.protocol_versions_two

        self._test_read(stream, payload, protocol_versions)

    def _test_write(self, payload, expected):
        stream = utils.BytearrayStream()
        payload.write(stream)

        length_expected = len(expected)
        length_received = len(stream)

        msg = "encoding lengths not equal"
        msg += "; expected {0}, received {1}".format(
            length_expected, length_received)
        self.assertEqual(length_expected, length_received, msg)

        msg = "encoding mismatch"
        msg += ";\nexpected:\n{0}\nreceived:\n{1}".format(expected, stream)

        self.assertEqual(expected, stream, msg)

    def test_write_with_empty_protocol_list(self):
        payload = payloads.DiscoverVersionsRequestPayload(
            self.protocol_versions_empty)
        expected = self.encoding_empty

        self._test_write(payload, expected)

    def test_write_with_one_protocol_version(self):
        payload = payloads.DiscoverVersionsRequestPayload(
            self.protocol_versions_one)
        expected = self.encoding_one

        self._test_write(payload, expected)

    def test_write_with_two_protocol_versions(self):
        payload = payloads.DiscoverVersionsRequestPayload(
            self.protocol_versions_two)
        expected = self.encoding_two

        self._test_write(payload, expected)


class TestDiscoverVersionsResponsePayload(TestCase):

    def setUp(self):
        super(TestDiscoverVersionsResponsePayload, self).setUp()

        self.protocol_versions_empty = list()
        self.protocol_versions_one = list()
        self.protocol_versions_one.append(ProtocolVersion(1, 0))
        self.protocol_versions_two = list()
        self.protocol_versions_two.append(ProtocolVersion(1, 1))
        self.protocol_versions_two.append(ProtocolVersion(1, 0))

        self.encoding_empty = utils.BytearrayStream((
            b'\x42\x00\x7C\x01\x00\x00\x00\x00'))
        self.encoding_one = utils.BytearrayStream((
            b'\x42\x00\x7C\x01\x00\x00\x00\x28\x42\x00\x69\x01\x00\x00\x00\x20'
            b'\x42\x00\x6A\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x6B\x02\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00'
            b'\x00'))
        self.encoding_two = utils.BytearrayStream((
            b'\x42\x00\x7C\x01\x00\x00\x00\x50\x42\x00\x69\x01\x00\x00\x00\x20'
            b'\x42\x00\x6A\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x6B\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x69\x01\x00\x00\x00\x20\x42\x00\x6A\x02\x00\x00\x00\x04'
            b'\x00\x00\x00\x01\x00\x00\x00\x00\x42\x00\x6B\x02\x00\x00\x00\x04'
            b'\x00\x00\x00\x00\x00\x00\x00\x00'))

    def tearDown(self):
        super(TestDiscoverVersionsResponsePayload, self).tearDown()

    def test_init_with_none(self):
        payloads.DiscoverVersionsResponsePayload()

    def test_init_with_args(self):
        payloads.DiscoverVersionsResponsePayload(
            self.protocol_versions_empty)

    def test_validate_with_invalid_protocol_versions(self):
        kwargs = {'protocol_versions': 'invalid'}
        self.assertRaisesRegex(
            TypeError, "invalid protocol versions list",
            payloads.DiscoverVersionsResponsePayload, **kwargs)

    def test_validate_with_invalid_protocol_version(self):
        kwargs = {'protocol_versions': ['invalid']}
        self.assertRaisesRegex(
            TypeError, "invalid protocol version",
            payloads.DiscoverVersionsResponsePayload, **kwargs)

    def _test_read(self, stream, payload, protocol_versions):
        payload.read(stream)
        expected = len(protocol_versions)
        observed = len(payload.protocol_versions)

        msg = "protocol versions list decoding mismatch"
        msg += "; expected {0} results, received {1}".format(
            expected, observed)
        self.assertEqual(expected, observed, msg)

        for i in xrange(len(protocol_versions)):
            expected = protocol_versions[i]
            observed = payload.protocol_versions[i]

            msg = "protocol version decoding mismatch"
            msg += "; expected {0}, received {1}".format(expected, observed)
            self.assertEqual(expected, observed, msg)

    def test_read_with_empty_protocol_list(self):
        stream = self.encoding_empty
        payload = payloads.DiscoverVersionsResponsePayload()
        protocol_versions = self.protocol_versions_empty

        self._test_read(stream, payload, protocol_versions)

    def test_read_with_one_protocol_version(self):
        stream = self.encoding_one
        payload = payloads.DiscoverVersionsResponsePayload()
        protocol_versions = self.protocol_versions_one

        self._test_read(stream, payload, protocol_versions)

    def test_read_with_two_protocol_versions(self):
        stream = self.encoding_two
        payload = payloads.DiscoverVersionsResponsePayload()
        protocol_versions = self.protocol_versions_two

        self._test_read(stream, payload, protocol_versions)

    def _test_write(self, payload, expected):
        stream = utils.BytearrayStream()
        payload.write(stream)

        length_expected = len(expected)
        length_received = len(stream)

        msg = "encoding lengths not equal"
        msg += "; expected {0}, received {1}".format(
            length_expected, length_received)
        self.assertEqual(length_expected, length_received, msg)

        msg = "encoding mismatch"
        msg += ";\nexpected:\n{0}\nreceived:\n{1}".format(expected, stream)

        self.assertEqual(expected, stream, msg)

    def test_write_with_empty_protocol_list(self):
        payload = payloads.DiscoverVersionsResponsePayload(
            self.protocol_versions_empty)
        expected = self.encoding_empty

        self._test_write(payload, expected)

    def test_write_with_one_protocol_version(self):
        payload = payloads.DiscoverVersionsResponsePayload(
            self.protocol_versions_one)
        expected = self.encoding_one

        self._test_write(payload, expected)

    def test_write_with_two_protocol_versions(self):
        payload = payloads.DiscoverVersionsResponsePayload(
            self.protocol_versions_two)
        expected = self.encoding_two

        self._test_write(payload, expected)
